#include <stdio.h>
#include <math.h>
/* for tcp/ip */
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h> /* for gethostbyname */
#include <sys/poll.h> /* for polling garbage*/


#include "quaternion.h"

#include "player.h"

#include "gameeng.h"
#include "glheightfield.h"

int game_netserver(game_t *gamedata);
void process_game(game_t *G);

int game_clear(game_t *gamedata) {
	int i;
	bullet_t *bptr, *prev_bptr;

        gamedata->bullet_list_end = NULL;
        gamedata->bullet_list_start = NULL;
	gamedata->loop_counter = 0;
	gamedata->max_packet_size = 0;
	gamedata->max_packet_b = 0;
	gamedata->max_packet_s = 0; 
	gamedata->max_packet_ps = 0; 
	gamedata->max_packet_pps = 0; 
	gamedata->loop_packets = 0;
	gamedata->num_bullets = 0;
	gamedata->loop_time = 0.0;
	gamedata->gravity = 0.0;
	gamedata->player_height = 0.0;
	gamedata->clamp_player_to_ground = 0;
        prev_bptr = NULL;

	/* gamedata->config_filename = NULL; */

        for (i = 0 ; i < 100 ; i++) { /* build list of 100 bullets */
                bptr = (bullet_t *) malloc(sizeof(bullet_t));
                bptr->next = prev_bptr;
                prev_bptr = bptr;
        }
        gamedata->free_bullets = bptr;
                                                                                                                                                          
        /* reset all players to be inactive, clear all client sockets */
        for (i = 0; i < GAME_MAX_CLIENTS ; i++) {
                player_clear(&gamedata->players[i]);
		gamedata->client_sockets[i] = 0;
		gamedata->client_game_cycle_delta[i] = 0;
	}
	return 0;
}


int game_connect(game_t *gamedata, char *game_server_name, int portno) {

	int sockfd;
        struct sockaddr_in server_addr; 
        struct hostent *he;

	if (portno == 0)
		portno = GAME_DEFAULT_PORT;

        if ((he = gethostbyname(game_server_name)) == NULL) {  
                perror("gethostbyname");
                exit(1);
	}
        if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
                perror("socket");
                exit(1);
        }
                                                                                
        server_addr.sin_family = AF_INET;         /* host byte order*/
        server_addr.sin_port = htons(portno); /* short, network byte order */
        server_addr.sin_addr = *((struct in_addr *)he->h_addr);
        memset(&(server_addr.sin_zero), '\0', 8);  /* zero the rest of the struct */
        /* connect to game port */
        if (connect(sockfd, (struct sockaddr *)&server_addr, sizeof(struct sockaddr))) {
        	perror("connect");
		close(sockfd);
		sockfd = -1;
	}
	gamedata->server_socket = sockfd;

	gamedata->servername = (char *) malloc(sizeof(char) * strlen(game_server_name));
	strcpy(gamedata->servername, game_server_name);
	gamedata->port = portno;
	game_clear(gamedata);
	printf("Game Engine Connect ip: %s port: %d. socket: %d\n", game_server_name, portno, sockfd);

	return(sockfd);
}

int game_join(game_t *gamedata, char *name) {  /* Add me to player list */
	game_packet_t join_packet;
	int plen;
	char bigbuffer[1024];
	player_t *pptr;

	join_packet.packet_type = GAME_JOIN;
	strncpy(join_packet.data.chardata, name, 28);
	printf("Game Engine join as <%s>\n", join_packet.data.chardata);
	game_send(gamedata, &join_packet);
	
	plen = game_recv(gamedata, bigbuffer, sizeof(player_t));
	if (plen != sizeof(player_t)) {
		fprintf(stderr, "Oh - no, packet size of %d not same as player_t! write more code here!\n", plen);
		exit(1);
	}
	
	pptr = (player_t *) bigbuffer;
	gamedata->first_person_index = pptr->player_id;
	player_copy(pptr, &gamedata->players[gamedata->first_person_index]);
	gamedata->players[gamedata->first_person_index].status = PLAYER_STATUS_ALIVE;
	gamedata->players[gamedata->first_person_index].player_name = malloc(sizeof(char) * 30); /* more than 28, the max */
	strncpy(gamedata->players[gamedata->first_person_index].player_name, name, 28);

	return 0;
}

int game_disconnect(game_t *gamedata) {
	printf("Game Engine disconenct\n");
	close(gamedata->server_socket);
	return(0);
}

int game_quit(game_t *gamedata) {
	game_packet_t kill_server_packet;

	printf("Game Engine Quit\n");
	kill_server_packet.packet_type = GAME_KILL;
	game_send(gamedata, &kill_server_packet);
	return(0);
}

int game_submit(game_t *gamedata, player_t *player) {
	game_packet_t p;

	p.packet_type = GAME_SUBMIT;
	p.data.intdata[0] = player->player_id;
	p.data.intdata[1] = player->key_actions;
	p.data.intdata[2] = player->mouse_buttons;
	p.data.intdata[3] = player->mouse_delta[0];
	p.data.intdata[4] = player->mouse_delta[1];
	game_send(gamedata, &p);
	
	return(0);
}

int game_update(game_t *gamedata) {
	game_packet_t *p;
	static game_packet_t ten_packs[101];
	int player_id;
	int done_transmission, update_packet_times;
	int active_players;
	int i, bytesgot, prev_bytesgot, numpacks, fracpack, totpak;
	bullet_t *bptr, *bptr2;
	struct pollfd mopupfd;
	stopwatch_t sw;

	gamedata->loop_packets = 0;
	if (gamedata->loop_counter % 500 == 0) { /* every 500'th update... */
		gamedata->max_packet_size = 0;
		gamedata->max_packet_wait = 0.0;
	}
	done_transmission = 0;
	active_players = 0;
	prev_bytesgot = 0;

	mopupfd.fd = gamedata->server_socket;
	mopupfd.events = POLLIN | POLLPRI;

	/* move bullet lists around */
	if (gamedata->bullet_list_end != NULL) {
		gamedata->bullet_list_end->next = gamedata->free_bullets;
		gamedata->free_bullets = gamedata->bullet_list_start;
		gamedata->bullet_list_start = NULL;
		gamedata->bullet_list_end = NULL;
	}
	/* now gamedata->free_bullets has all the bullets, start and end = NULL */
	bptr = gamedata->free_bullets;

	/* mark all players as inactive */
	for (i = 0; i < GAME_MAX_CLIENTS; i++)
		gamedata->players[i].status = PLAYER_STATUS_INACTIVE;

	numpacks = bytesgot = fracpack = 0;
	totpak = 0;
	while (! done_transmission ) {
		totpak++;
		update_packet_times = 0;
		if (totpak != 1) {
			stopwatch(&sw, 0);
			if (poll(&mopupfd, 1, 800) == 0) {
				printf("Socket dried up after %d packets, last stats: bytesgot: %d, numpacks: %d, frac: %d\n",
					totpak, bytesgot, numpacks, fracpack);
				done_transmission = 3;
				bytesgot = 0;
				fracpack = 0;
			}
			if (gamedata->max_packet_wait < stopwatch(&sw, 1)) {
				gamedata->max_packet_wait = stopwatch(&sw, 1);
				gamedata->max_packet_b = totpak;
				update_packet_times = 1;
				gamedata->max_packet_ps = bytesgot;
				gamedata->max_packet_pps = prev_bytesgot;
			}
		} 

		if (done_transmission != 3) {
			if (fracpack != 0)  /* move data into position...*/
				memmove(ten_packs, &ten_packs[numpacks], fracpack);
			prev_bytesgot = bytesgot;
			bytesgot = game_recv(gamedata,(char *) ten_packs + fracpack, sizeof(game_packet_t) * 100);
		}

		if (bytesgot > gamedata->max_packet_size)
			gamedata->max_packet_size = bytesgot;
		numpacks = ( bytesgot + fracpack) / sizeof(game_packet_t);
		fracpack = ( bytesgot + fracpack) % sizeof(game_packet_t);
		/*if (fracpack) {
			printf("Got %dth packet of %d bytes.  That's %d and %d/%d-iths packets\n", totpak, bytesgot, numpacks, fracpack, sizeof(game_packet_t));
			
		} */
		if (update_packet_times == 1) {
			gamedata->max_packet_s = bytesgot;
		}
		for (i = 0; i < numpacks ; i++) {
			p = &ten_packs[i];	
			switch (p->packet_type) {
				case GAME_PLAYER_POSITION:
					player_id = p->data.intdata[5];
					gamedata->players[player_id].status      = p->data.intdata[6];
					gamedata->players[player_id].location[0] = p->data.floatdata[0];
					gamedata->players[player_id].location[1] = p->data.floatdata[1];
					gamedata->players[player_id].location[2] = p->data.floatdata[2];
					gamedata->players[player_id].heading[0]  = p->data.floatdata[3];
					gamedata->players[player_id].heading[1]  = p->data.floatdata[4];
					active_players |= 1 << player_id; /* tag as active player */
					break;
				case GAME_PLAYER_NAME: /* 5 floats and an int. */
					player_id = p->data.intdata[5];
					gamedata->players[player_id].player_name = (char *) malloc(sizeof(char) * strlen(p->data.chardata));
					strcpy( gamedata->players[player_id].player_name , p->data.chardata);
					break;
				case GAME_PLAYER_EVENT:
					break;
				case GAME_BULLET_POSITION:
					if (bptr == NULL) { 
						bptr = (bullet_t *) malloc(sizeof(bullet_t));
						bptr->next = NULL;
					} 
					bptr->location[0] =  p->data.floatdata[0];
					bptr->location[1] =  p->data.floatdata[1];
					bptr->location[2] =  p->data.floatdata[2];
					bptr->heading[0]  =  p->data.floatdata[3];
					bptr->heading[1]  =  p->data.floatdata[4];
					bptr->type 	  =  p->data.intdata[5];
					if (gamedata->bullet_list_end == NULL) {
						gamedata->bullet_list_end = bptr;
						gamedata->bullet_list_start = bptr;
						bptr = bptr->next; 
						gamedata->bullet_list_end->next = NULL;
					} else {
						bptr2 = gamedata->bullet_list_start;
						gamedata->bullet_list_start = bptr;
						bptr = bptr->next; 
						gamedata->bullet_list_start->next = bptr2;
					}
					break;
				case GAME_BULLET_EVENT:
					break;
				case GAME_UPDATE_END:
					gamedata->players[gamedata->first_person_index].hit_points = p->data.intdata[0];
					gamedata->client_game_cycle_delta[gamedata->first_person_index] = p->data.intdata[1];
					done_transmission = 1;
					break;
				default:
					fprintf(stderr, "Not ready for packet type %d.  Happened after %d packets, latest %d of %d w/ %d frac\n", 
						p->packet_type, totpak, i, numpacks, fracpack);
					/* consider update as garbage. say we're done */
					done_transmission = 2;
					break;
				} /* switch packet type */
			} /* for packets */
		} /* while not done transmitting */

	/* if I updated bullets, fix free list */
	if (bptr != gamedata->free_bullets) {
		gamedata->free_bullets = bptr;
	}

	/* if transmission is done, could be an error.  Mop up socket mess */
	if (done_transmission == 2) {
		printf("mopping up socket\n");
		while (poll(&mopupfd, 1, 1)) /* suck socket dry */
 			bytesgot = game_recv(gamedata,(char *) ten_packs, sizeof(game_packet_t) * 10);
	}

	/* if I updated players, mark non-updates as inactive */
	/*if (active_players)
		for (i = 0 ; i < GAME_MAX_CLIENTS ; i++) 
			if (! (active_players &  (1 << i)))
				gamedata->players[i].status = PLAYER_STATUS_INACTIVE;
				
*/
	gamedata->loop_packets = totpak;
	return(0);
}

int game_send(game_t *gamedata, game_packet_t *packet) {
	/* ahh, the pleasure of abstraction */
	int total = 0;      
        int bytesleft;
        int n;


	bytesleft = sizeof(game_packet_t);
	
	while ( total < sizeof(packet)) {
 		n = send(gamedata->server_socket, packet + total, bytesleft, 0);
		if (n == -1) {break;} /* error in send */
		total += n;
		bytesleft -=n;
	}
	if (n == -1) /* bailed on error */	
		perror("send");
	return(n);
}

int client_send(int client_socket, char *buf, int buflen) {
	int total = 0;      
        int bytesleft;
        int n = 0;
	
	bytesleft = buflen;
	while ( total < buflen) {
		if (total != 0) 
			printf("had to resend after %d bytes of total %d bytes\n", total, buflen);
 		n = send(client_socket, buf + total, bytesleft, 0);
		if (n == -1) {break;} /* error in send */
		total += n;
		bytesleft -=n;
	}

/*	send(client_socket, buf, buflen, 0);*/
	if (n == -1) /* bailed on error */	
		perror("send");
	return(total);
}

int clients_send(game_t *gamedata, char *buf, int buflen) {
	int i, clientfd;

	for (i = 0 ; i < GAME_MAX_CLIENTS; i++) {
		clientfd = gamedata->client_sockets[i];
		if (clientfd != 0 && gamedata->client_updated_me[i] >= 1) 
					client_send(clientfd, buf, buflen);
	}
	return 0;
}
	

#define BIG_PACKET_SIZE 128
int clients_update(game_t *gamedata) {
	int i, j, clientfd, big_packet_index;
	player_t *pp;
	bullet_t *bp;
	game_packet_t *gp; 
	static game_packet_t big_packet[BIG_PACKET_SIZE];
	
	/* new: compose big_packet and send to everyone at once.  Don't
		have to compose packet for every client then */
	big_packet_index = 0;

	/* send player data */
	for (j = 0 ; j < GAME_MAX_CLIENTS; j++) {
		gp = &big_packet[big_packet_index];
		pp = &gamedata->players[j];
		if (pp->status != PLAYER_STATUS_INACTIVE) {
			gp->packet_type = GAME_PLAYER_POSITION;

			gp->data.floatdata[0] = pp->location[0];
			gp->data.floatdata[1] = pp->location[1];
			gp->data.floatdata[2] = pp->location[2];
			gp->data.floatdata[3] = pp->heading[0];
			gp->data.floatdata[4] = pp->heading[1];
			gp->data.intdata[5]   = pp->player_id;
			gp->data.intdata[6]   = pp->status;


			if (big_packet_index == (BIG_PACKET_SIZE - 1)) {
				clients_send(gamedata, (char *) big_packet,
					sizeof(game_packet_t) * BIG_PACKET_SIZE);
				big_packet_index = 0;
			} else {
				big_packet_index++;
			}
		}
	} /* send players */

	/* send bullets */
	bp = gamedata->bullet_list_start;
	while (bp) {
		gp = &big_packet[big_packet_index];
		gp->packet_type = GAME_BULLET_POSITION;
		gp->data.floatdata[0] = bp->location[0];
		gp->data.floatdata[1] = bp->location[1];
		gp->data.floatdata[2] = bp->location[2];
		gp->data.floatdata[3] = bp->heading[0];
		gp->data.floatdata[4] = bp->heading[1];
		gp->data.intdata[5]   = bp->type;
		if (big_packet_index == (BIG_PACKET_SIZE - 1)) {
			clients_send(gamedata, (char *) big_packet,
				sizeof(game_packet_t) * BIG_PACKET_SIZE);
			big_packet_index = 0;
		} else {
			big_packet_index++;
		}
		bp = bp->next;
	} /* send bullets */
		
	/* now send last packet to everyone with individual data */
	gp = &big_packet[big_packet_index];
	for (i = 0 ; i < GAME_MAX_CLIENTS; i++) {
		clientfd = gamedata->client_sockets[i];
		if (clientfd != 0 && gamedata->client_updated_me[i] >= 1) { 
			gp->packet_type = GAME_UPDATE_END; 
			/* hit points */
			gp->data.intdata[0] = gamedata->players[i].hit_points;
			/* cycles since last submission */
			gp->data.intdata[1] = gamedata->client_game_cycle_delta[i];
			gamedata->client_game_cycle_delta[i] = 0;
			client_send(clientfd, (char *) big_packet,
				sizeof(game_packet_t) * ( big_packet_index + 1));
		} /* if client updated me */
	}
	return(0);
}

int game_recv(game_t *gamedata, char *buffer, int bufsize) {
	return (recv(gamedata->server_socket, buffer, bufsize, 0));
}

int game_init(game_t *gamedata) {
	int mypid;

	mypid = fork();

	if (mypid == 0) { /* I'm the child */
		printf("Child game file name %s\n", gamedata->config_filename);
		game_fileconfig(gamedata);
		game_netserver(gamedata);
		exit(0);
	}
	else if (mypid == -1) { /* failed to fork */
		fprintf(stderr, "Failed to fork net server\n");
		exit(1);
	}

	printf("Parent game file name %s\n", gamedata->config_filename);
	/* return pid of child */
	return mypid;
}

void client_hangup(int nbytes, int socketno, fd_set *masterp) {

	if (nbytes == 0) {
		/* connection closed */
		printf("selectserver: socket %d hung up\n", socketno);
	} else {
		perror("recv");
	}
	close(socketno); /* bye! */
	FD_CLR(socketno, masterp); /* remove from master set */
}

int game_parse_net_packet(game_t *gamedata, char *buf, int nbytes,  game_packet_t *game_packets, int client_file_id) {
	game_packet_t *tmp;
	player_t *play_ptr;
	int i, numpacks, fracpack, packloop, server_instruction, prev_type;

	numpacks = nbytes / sizeof(game_packet_t);
	fracpack = nbytes % sizeof(game_packet_t);
	if (fracpack)
		printf("Got packet of %d bytes.  That's %d and %d/%d-iths packets\n", nbytes, numpacks, fracpack, sizeof(game_packet_t));
	

	server_instruction = 0;
	/* printf("\n");
	for ( i = 0; i < nbytes ; i++)
		printf("%2x ", (unsigned char) buf[i]);
	printf("\n"); */
	tmp = NULL;
	prev_type = 0;
	for (packloop = 0 ; packloop < numpacks; packloop++) {
		if (tmp)
			prev_type = tmp->packet_type;
		tmp = (game_packet_t *) &buf[packloop * sizeof(game_packet_t)];
		if (packloop == 1) {
			printf("Got more than one packet from %d! first was %d then %d\n",
				client_file_id, prev_type, tmp->packet_type);
		}
		switch(tmp->packet_type) {
			case 0:
				printf("Null\n");
				break;	
			case 1:
				printf("Ints: ");
				for (i = 0 ; i < 7 ; i++)
					printf("%d ", (int) tmp->data.intdata[i]);
				printf("\n");
				break;	
			case 2:
				printf("Floats: ");
				for (i = 0 ; i < 7 ; i++)
					printf("%3.2f ", tmp->data.floatdata[i]);
				printf("\n");
				break;	
			case 3:
				printf("Char: %s\n", tmp->data.chardata);
				break;	
			case GAME_KILL: /* kill server */
				server_instruction = GAME_KILL;
				break;
			case 99: /* location set */
				printf("Location set: %3.2f %3.2f %3.2f\n", 
						tmp->data.floatdata[0], 
						tmp->data.floatdata[1], 
						tmp->data.floatdata[2]);
				break;
			case GAME_JOIN: /* join game , get player id */
				printf("join game name: %s\n", tmp->data.chardata);
				gamedata->client_sockets[client_file_id % GAME_MAX_CLIENTS] = client_file_id; /* save socket */
				play_ptr = &gamedata->players[client_file_id % GAME_MAX_CLIENTS];
				player_clear(play_ptr);
				play_ptr->player_id = client_file_id % GAME_MAX_CLIENTS;
				play_ptr->player_name = (char *) malloc(sizeof(char) * strlen(tmp->data.chardata));
				strcpy(play_ptr->player_name, tmp->data.chardata);
				play_ptr->location[0] = (float) client_file_id * 10.0;	
				play_ptr->location[1] = 1.0;
				play_ptr->location[2] = 0.0;
				play_ptr->heading[0] = 180.0;
				play_ptr->heading[1] = 0.0;
				play_ptr->mouse_buttons = 0x00;
				play_ptr->key_actions = 0;
				play_ptr->velocity = 0.15;
				play_ptr->status = PLAYER_STATUS_ALIVE;
				play_ptr->height = gamedata->player_height;
				player_calc_orientation(play_ptr);
				client_send(client_file_id, (char *) play_ptr, sizeof(player_t));
				break;
			
			case GAME_SUBMIT: 
				if (tmp->data.intdata[0] != client_file_id % GAME_MAX_CLIENTS) {
					fprintf(stderr, "YoW! indexes don't match!\n");
					fprintf(stderr, "YoW! packet: %d vs client %d\n",
				tmp->data.intdata[0] ,client_file_id % GAME_MAX_CLIENTS);
				}
			        play_ptr = &gamedata->players[tmp->data.intdata[0]]; /* who are you? */
			        play_ptr->key_actions    = tmp->data.intdata[1];
        			play_ptr->mouse_buttons  = tmp->data.intdata[2];
        			play_ptr->mouse_delta[0] = tmp->data.intdata[3];
        			play_ptr->mouse_delta[1] = tmp->data.intdata[4];
				gamedata->client_updated_me[client_file_id] += 1;
				break;

			default:
				printf("Unknown packet type: %d\n", tmp->packet_type);
				break;	
		}
	}
	return server_instruction;
}
	
int game_netserver(game_t *game) {
	fd_set master;   /* master file descriptor list */
	fd_set read_fds; /* temp file descriptor list for select() */
	struct sockaddr_in myaddr;	 /* server address */
	struct sockaddr_in remoteaddr; /* client address */
	int fdmax;		/* maximum file descriptor number */
	int listener;	 /* listening socket descriptor */
	int newfd;		/* newly accept()ed socket descriptor */
	char buf[256];	/* buffer for client data */
	int nbytes;
	int yes=1;		/* for setsockopt() SO_REUSEADDR, below */
	unsigned int addrlen;
	int i, server_cmd, portno;
	game_packet_t	in_packet;
	stopwatch_t sw;
	
	game_clear(game); /* reset player list / allocate bullets */
	game_fileconfig(game);	/* configure game variables via config file */
	make_height_game(&game->heightfield);

	portno = game->port;
	if (portno == 0)
		portno = GAME_DEFAULT_PORT;
	printf("Game Engine Init, using port %d.\n", portno);

	FD_ZERO(&master);	/* clear the master and temp sets */
	FD_ZERO(&read_fds);

	/* get the listener */
	if ((listener = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
		perror("socket");
		exit(1);
	}

	game->server_socket = listener;

	/* lose the pesky "address already in use" error message */
	if (setsockopt(listener, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1) {
		perror("setsockopt");
		exit(1);
	}

	/* bind */
	myaddr.sin_family = AF_INET;
	myaddr.sin_addr.s_addr = INADDR_ANY;
	myaddr.sin_port = htons(portno);
	memset(&(myaddr.sin_zero), '\0', 8);
	if (bind(listener, (struct sockaddr *)&myaddr, sizeof(myaddr)) == -1) {
		perror("bind");
		exit(1);
	}

	/* listen */
	if (listen(listener, GAME_MAX_CLIENTS) == -1) {
		perror("listen");
		exit(1);
	}

	/* add the listener to the master set */
	FD_SET(listener, &master);

	/* keep track of the biggest file descriptor */
	fdmax = listener; /* so far, it's this one */

	stopwatch(&sw, 0); /*reset stopwatch */
	/* main loop */
	for(;;) {
		game->loop_counter++;
		game->loop_time = stopwatch(&sw, 2);
		/* disable real time based velocites: game->loop_time = .03; */
		if (game->loop_counter % 1220 == 0)  { /* report stats */
			printf("Cycle rate: %3.2f per Second. Processed %d bullets\n", 10.0 / (game->loop_time + 0.000001),  game->num_bullets);
		}
		read_fds = master; /* copy it */

		/* clear client-updated-me table */
		for (i = 0 ; i < GAME_MAX_CLIENTS ; i++)
			game->client_updated_me[i] = 0;
		
		if (select(fdmax+1, &read_fds, NULL, NULL, NULL) == -1) {
			perror("select");
			exit(1);
		}

		/* run through the existing connections looking for data to read */
		for(i = 0; i <= fdmax; i++) {
			if (FD_ISSET(i, &read_fds)) { /* we got one!! */
				if (i == listener) {
					/* handle new connections */
					addrlen = sizeof(remoteaddr);
					if ((newfd = accept(listener, (struct sockaddr *)&remoteaddr, &addrlen)) == -1) { 
						perror("accept");
					} else {
						FD_SET(newfd, &master); /* add to master set */
						if (newfd > fdmax) {	/* keep track of the maximum */
							fdmax = newfd;
						}
						printf("selectserver: new connection from %s on "
							"socket %d\n", inet_ntoa(remoteaddr.sin_addr), newfd);
					}
				} else {
					/* handle data from a client */
					nbytes = recv(i, buf, sizeof(buf), 0);
					if (nbytes <= 0) { /* got error or connection closed by client */
						client_hangup(nbytes, i, &master);
						game->client_sockets[i] = 0;
                				game->players[i].status = PLAYER_STATUS_INACTIVE;
					} else { /* we got some data from a client */
						if (nbytes == sizeof(buf)) {
							printf("GRGGG! choking on too many packets from %d!\n", i);
						}
						server_cmd = game_parse_net_packet(game, buf, nbytes, &in_packet, i);
						if (server_cmd == GAME_KILL) {  /* kill server */
							printf("Got kill server command, exiting...\n");
							exit(0);
						}
					} /* if nbytes */
				} /* if listener */
			} /* if fdis set */
		} /* for i fdmax */
		/* process game here! */
		
		process_game(game);
		/* send updated game to clients */
		for (i = 0 ; i < GAME_MAX_CLIENTS ; i++)
			game->client_game_cycle_delta[i]++;
		clients_update(game);
	}
	
	return 0;
}

void playerUpdate(game_t *G) {
	int i;
	player_t *p;

	for (i = 0 ; i < GAME_MAX_CLIENTS ; i++) {
		p = &G->players[i];
		switch(p->status) {
			case PLAYER_STATUS_DEAD: /* bring back from the dead */
				p->status = PLAYER_STATUS_ALIVE;
				p->hit_points = 100;
				p->location[0] = (float)( G->loop_counter % 100);
				p->location[2] =  (float) (( G->loop_counter * 81) % 100);
				break;
			default:
				break;
		}
	}
}
				
	
void bulletUpdate(game_t *G) {
	int i, delete_bullet;
	bullet_t *b, *prev_b, *next_b;
	float bullet_dist;
	float velocity_factor;

	b = G->bullet_list_start;
	G->num_bullets = 0;
	prev_b = NULL;

	velocity_factor = G->loop_time * 10.0; /* seconds */

	while (b) {
		delete_bullet = 0;
		G->num_bullets++;
		next_b = b->next;

		/* move bullet */
		bullet_move(b, velocity_factor);

		switch (b->type) {
			case BULLET_TYPE_ROCKET:
				/* check for radius and remove */
				bullet_dist = bullet_length_2(b);
				if (bullet_dist > 100000.0) { /* that's 1000 units */
					delete_bullet = 1;
				} else {
					if (bullet_dist > 9.0) { /* let it leave the barrel before testing for hits */
						/* check for player hit */
						for (i = 0 ; i < GAME_MAX_CLIENTS ; i++) {
							if (G->players[i].status == PLAYER_STATUS_ALIVE)
								if (bullet_hit_player(b, &G->players[i]) > 0) ;
									/* delete_bullet = 1; */
						}
						/* check for ground hit */
						if (detect_below(&G->heightfield, b->location) >= 0.0) {
							b->type = BULLET_TYPE_EXPLOSION;
                					b->heading[0] = 0.1;
						}
					}	
				}
				break;
			case BULLET_TYPE_EXPLOSION:
				/* check if heading[0] > 1.0 */
				if (b->heading[0] > 5.0)
					delete_bullet = 1;
				break;
			case BULLET_TYPE_SHIP_EXPLOSION:
				/* check if heading[0] > 1.0 */
				if (b->heading[0] > 35.0)
					delete_bullet = 1;
				break;
		}

		if (delete_bullet == 1) {
			/* check if this is front of bullet_list_start */
			if (prev_b == NULL)
				G->bullet_list_start = b->next;
			else
				prev_b->next = b->next;

			/* tack to front of free list */
			b->next = G->free_bullets;
			G->free_bullets = b;
		} else {
			prev_b = b;
		}

		b = next_b;
		/* migrate to explosion, etc. */
	}
}

void keyAndMouseUpdate(game_t *G, player_t *p) {
	float cosxz, sinxz, rtc, rts, upness, cosupness, hf_elev;
	int   kbd, mx, my, mb;
	int 	heading_changed;
	bullet_t *bptr;
	float velocity_factor, gravity_factor;

	kbd = p->key_actions;
	mb = p->mouse_buttons;
	mx = p->mouse_delta[0];
	my = p->mouse_delta[1];
	/* zero deltas in case client misses update. */
	p->mouse_delta[0] = p->mouse_delta[1] = 0;
	heading_changed = 0;

	/* skip sin/cos calls if no physical change. */
	bptr = G->free_bullets;
	
	if ((kbd | mb | mx | my) != 0) {
		velocity_factor = G->loop_time * 10.0; /* seconds */



		if ((mx | my) != 0) {
			p->heading[0] -= (float) (my) * GAME_MOUSE_ACCEL; /* later player's accel */
			p->heading[1] += (float) (mx) * GAME_MOUSE_ACCEL; /* later player's accel */
			/* clamp angles */
			if (p->heading[0] > 89.0) p->heading[0] = 89.0;	
			if (p->heading[0] < -89.0) p->heading[0] = -89.0;	
			if (p->heading[1] > 360.0) p->heading[1] -= 360.0;	 
			if (p->heading[1] < 0.0) p->heading[1] += 360.0;	
			heading_changed = 1;
		}
		if (kbd & PLAYER_KEY_LEFT)  {
                        p->heading[1] -= 1.4; 
			heading_changed = 1;
		}
		if (kbd & PLAYER_KEY_RIGHT) {
                        p->heading[1] += 1.4;
			heading_changed = 1;
		}
		
		if (heading_changed) { /* recalculate angles */
			player_calc_orientation(p);
		}
		cosxz = p->cosxz;
        	sinxz = p->sinxz;
		rtc   = p->rtc * velocity_factor;
		rts   = p->rts * velocity_factor;
        	cosupness = p->cosupness * velocity_factor;
        	upness =    p->upness * velocity_factor;

		if (kbd & PLAYER_KEY_FORWARD) {
                        p->location[2] -= cosxz * cosupness;
                        p->location[1] -= upness;
                        p->location[0] -= sinxz * cosupness; 
		}
		if (kbd & PLAYER_KEY_BACKWARD) {
                        p->location[2] += cosxz * cosupness;
                        p->location[1] += upness;
                        p->location[0] += sinxz * cosupness; 
		}

		if (kbd & PLAYER_KEY_STRAFE_LEFT)  {
                        p->location[0] -= rts;
			p->location[2] -= rtc; 
		}
		if (kbd & PLAYER_KEY_STRAFE_RIGHT)  {
                        p->location[0] += rts;
			p->location[2] += rtc; 
		}
		/* clamp player to ground */


		if (mb & PLAYER_MOUSE_BUTTON_0) { /*shoot!*/
			if (bptr == NULL) { /* no free bullets, so alloc */
				bptr = (bullet_t *) malloc(sizeof(bullet_t));
				bptr->next = NULL;
			}
			player_shoot(p, bptr);
			/* remove from free bullet list */
			G->free_bullets = bptr->next;
			/* tack to front of bullet list */
			bptr->next = G->bullet_list_start;
			G->bullet_list_start = bptr;
			/* server does not use end of list */
			/* advance to next free bullet */
			bptr = G->free_bullets;
		}
	} /* if moved */

	if (G->gravity > 0.0) {
		gravity_factor = G->loop_time * 10.0 * G->gravity; /* seconds */
		p->location[1] -= gravity_factor;
	}
		
        if (G->clamp_player_to_ground) {
		hf_elev = detect_below(&G->heightfield, p->location);
		if (hf_elev >= 0.0)
			p->location[1] = hf_elev;
	}	
}


/* process key, mouse moves for all players.
initiate bullets.  process bullet collisions.
etc. */
void process_game(game_t *G)
{
	int i;

	/* move everyone around */
	for (i = 0 ; i < GAME_MAX_CLIENTS; i++) {
		if (G->players[i].status == PLAYER_STATUS_ALIVE) {
			keyAndMouseUpdate(G, &G->players[i]);	
		}
	}

	/* check for ouches */
	bulletUpdate(G);

	/* check for death, explosions */
	playerUpdate(G);
}

int game_fileconfig(game_t *gamedata) {	/* configure game variables via config file */
	FILE *infile;
	char inbuf[256];
	int numscanned;
	int valid_option;
	
	if (gamedata->config_filename != NULL) {
		infile = fopen(gamedata->config_filename, "r");
		if (! infile) {
			fprintf(stderr, "Could not open file %s for reading config. Aborting...\n",
				gamedata->config_filename);
			exit(1);
		}
		while ( fgets(inbuf, 256, infile)) {
			valid_option = 0;
			if (strncmp(inbuf, "heightfield", strlen("heightfield")) == 0) { 
				gamedata->heightfield.heightfname = (char *) malloc(sizeof(char) * 256);
				gamedata->heightfield.texfname = (char *) malloc(sizeof(char) * 256);
				numscanned = sscanf(inbuf, "heightfield location %f %f %f scale %f %f %f heightfile %s texfile %s",
					&gamedata->heightfield.location[0],
					&gamedata->heightfield.location[1],
					&gamedata->heightfield.location[2],
					&gamedata->heightfield.scale_factors[0],
					&gamedata->heightfield.scale_factors[1],
					&gamedata->heightfield.scale_factors[2],
					gamedata->heightfield.heightfname,
					gamedata->heightfield.texfname);
				if (numscanned != 8) {
					fprintf(stderr, "Could not parse heightfiled line: \n%s\n in file %s config. Aborting...\n",
						inbuf,
						gamedata->config_filename);
					exit(1);
				}
				valid_option = 1;
			} /* if heightfiled config line */
			if (strncmp(inbuf, "servername", strlen("servername")) == 0) { 
				gamedata->servername = (char *) malloc(sizeof(char) * 256);
				numscanned = sscanf(inbuf, "servername %s",
					gamedata->servername);
				if (numscanned != 1) {
					fprintf(stderr, "Could not parse servername line: \n%s\n in file %s config. Aborting...\n",
						inbuf,
						gamedata->config_filename);
					exit(1);
				}
				valid_option = 1;
			}
			if (strncmp(inbuf, "port", strlen("port")) == 0) { 
				gamedata->servername = (char *) malloc(sizeof(char) * 256);
				numscanned = sscanf(inbuf, "port %d",
					&gamedata->port);
				if (numscanned != 1) {
					fprintf(stderr, "Could not parse port line: \n%s\n in file %s config. Aborting...\n",
						inbuf,
						gamedata->config_filename);
					exit(1);
				}
				valid_option = 1;
			}
			if (strncmp(inbuf, "ground_player", strlen("ground_player")) == 0) { 
				gamedata->servername = (char *) malloc(sizeof(char) * 256);
				numscanned = sscanf(inbuf, "ground_player %d",
					&gamedata->clamp_player_to_ground);
				if (numscanned != 1) {
					fprintf(stderr, "Could not parse ground_player line: \n%s\n in file %s config. Aborting...\n",
						inbuf,
						gamedata->config_filename);
					exit(1);
				}
				valid_option = 1;
			}
			if (strncmp(inbuf, "gravity", strlen("gravity")) == 0) { 
				gamedata->servername = (char *) malloc(sizeof(char) * 256);
				numscanned = sscanf(inbuf, "gravity %f",
					&gamedata->gravity);
				if (numscanned != 1) {
					fprintf(stderr, "Could not parse gravity line: \n%s\n in file %s config. Aborting...\n",
						inbuf,
						gamedata->config_filename);
					exit(1);
				}
				valid_option = 1;
			}
			if (strncmp(inbuf, "player_height", strlen("player_height")) == 0) { 
				gamedata->servername = (char *) malloc(sizeof(char) * 256);
				numscanned = sscanf(inbuf, "player_height %f",
					&gamedata->player_height);
				if (numscanned != 1) {
					fprintf(stderr, "Could not parse player_height line: \n%s\n in file %s config. Aborting...\n",
						inbuf,
						gamedata->config_filename);
					exit(1);
				}
				valid_option = 1;
			}
			/* comments in config file start with a hash: */
			if (*inbuf == '#') {
				valid_option = 1;
			}
			if ( valid_option == 0) {
				fprintf(stderr, "Could not parse line: \n%s\n in file %s config. Ignoring...\n",
					inbuf,
					gamedata->config_filename);
			}
			
		} /* while reading file by line */
	} /* if I have a config file name */
	return 1;
}

			

float stopwatch(stopwatch_t *sw, int function) {

        struct timeval tv;
                                                                                                                                  
        float computed_timeshift;
        time_t elapsed_secs, elapsed_usecs;
                                                                                                                                  
        gettimeofday(&tv, NULL);

        switch (function) {
                case 0: /* reset */
                        sw->reset_time.tv_sec  = sw->query_time.tv_sec  = tv.tv_sec;
                        sw->reset_time.tv_usec = sw->query_time.tv_usec = tv.tv_usec;
                        elapsed_secs = 0;
                        elapsed_usecs = 0;
                        break;
                case 1: /* query time since reset */
                        elapsed_secs = tv.tv_sec - sw->reset_time.tv_sec;
                        elapsed_usecs = tv.tv_usec - sw->reset_time.tv_usec;
                        sw->query_time.tv_sec  = tv.tv_sec;
                        sw->query_time.tv_usec = tv.tv_usec;
                        break;
                case 2: /* query time since last query */
                        elapsed_secs = tv.tv_sec - sw->query_time.tv_sec;
                        elapsed_usecs = tv.tv_usec - sw->query_time.tv_usec;
                        sw->query_time.tv_sec  = tv.tv_sec;
                        sw->query_time.tv_usec = tv.tv_usec;
                        break;
                default:
                        elapsed_secs = 0;
                        elapsed_usecs = 0;
                        fprintf(stderr, "No stopwatch function %d\n", function);
        }
                                                                                                                                  
        /* tenths of a second */
        computed_timeshift = (float) (elapsed_secs * 1000000.0 + elapsed_usecs)/
                                                      100000.0;
        return(computed_timeshift);
}

