#include <stdio.h>
#include <stdlib.h>
#include <GL/gl.h>
#include <GL/glu.h>

#include "ppmgl.h"
#include "glheightfield.h"

void make_height_heights(glheight_t *H) {
        GLubyte *imagepixels;
        int xres, yres, xc, yc, pindex;
	float *fp, one_twofiftysixth_scaled, hzscale, hzbase, hzmax;


        imagepixels = get_ppm_pixels(H->heightfname, &xres, &yres, 4, 0.0);
	H->xdim = xres;
	H->ydim = yres;
        H->heights = (float *) malloc(sizeof(float)* xres * yres);

	H->bbox[0] = H->location[0] + H->scale_factors[0];
	H->bbox[2] = H->location[2] + H->scale_factors[2];


	fp = H->heights;
	hzscale = H->scale_factors[1];
	hzbase = H->location[1];
	hzmax = 0.0;
	one_twofiftysixth_scaled = ( 1.0 / 256.0) * hzscale;
	for ( yc = 0 ; yc < yres ; yc++) {
       		for ( xc = 0 ; xc < xres; xc++) {
			pindex = ( yc * xres + xc ) * 4;
			*fp =  (float) imagepixels[pindex] * one_twofiftysixth_scaled + hzbase;
			if (*fp > hzmax)
				hzmax = *fp;
			fp++;
		}
	}
	H->bbox[1] = hzmax;
}


	
void make_height_texture(glheight_t *H) {
	makePPMTexture(H->texfname);
}

void make_height_game(glheight_t *H) {
	make_height_heights(H);
}

void make_height_gr(glheight_t *H) {

	int xres, yres, xc, yc, pindex;
	float xi, yi, xd, yd, z1, z2, o1[3], o2[3], vx[3], vy[3], n1[3], n2[3];

	make_height_heights(H);

	xres = H->xdim;
	yres = H->ydim;
	xd = H->scale_factors[0] / (GLfloat) xres;
        yd = H->scale_factors[2] / (GLfloat) yres;

	H->dlist = glGenLists(1);
	glNewList(H->dlist, GL_COMPILE);

	for ( yc = 0 ; yc < (yres - 1); yc++) {
		glBegin(GL_QUAD_STRIP);
		for ( xc = 0 ; xc < xres; xc++) {
			xi = (float) xc * xd;
			yi = (float) yc * yd;
			pindex =  yc * xres + xc;
			z1 = H->heights[pindex];
			pindex =  (yc + 1) * xres + xc;
			z2 = H->heights[pindex];
			/* have two points now.  Calc normals next */
			/* ADDME: Normals */
			glTexCoord2f( (GLfloat) xc / (GLfloat) xres, (GLfloat) yc / (GLfloat) yres);
                        glVertex3f( xi, z1, yi);
			glTexCoord2f( (GLfloat) xc / (GLfloat) xres, (GLfloat) ( yc + 1) / (GLfloat) yres);
                        glVertex3f( xi, z2, yi + yd);
		}
		glEnd();
	}
	glEndList();
	glGenTextures(1, &H->texid);
	glBindTexture(GL_TEXTURE_2D, H->texid);
	/* texture parameters */
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE); 


	make_height_texture(H);
}

void draw_height_gr(glheight_t *H) {
	glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, H->texid); 
	glCallList(H->dlist);
	glDisable(GL_TEXTURE_2D);
}

float get_height_at_point(glheight_t *H, float *loc) {
        float tmp, fi, fj, h00, h01, h10, h11, tmp1, tmp2;
        int i, j;

	/* check if point *loc is in bounding rectangle of heightfield */

        if ((loc[0] < H->bbox[0]) && (loc[0] > H->location[0]) && (loc[2] < H->bbox[2]) && (loc[2] > H->location[2])) {
                tmp = ( loc[0] - H->location[0] ) / (H->bbox[0] - H->location[0]);
                i = tmp * H->xdim;
                fi = tmp * (float) H->xdim - i;
                tmp = ( loc[2] - H->location[2] ) / (H->bbox[2] - H->location[2]);
                j = tmp * H->ydim;
                fj = tmp * (float) H->ydim - j;
                h00 = H->heights[j * H->xdim + i];
                h01 = H->heights[j * H->xdim + i + 1];
                h10 = H->heights[(j+1) * H->xdim + i];
                h11 = H->heights[(j+1) * H->xdim + i + 1];
                /* bilinear interpolate height */
                tmp1 = h00 * (1.0 - fi ) + h01 * fi;
                tmp2 = h10 * (1.0 - fi ) + h11 * fi;
                tmp = tmp1 * (1.0 - fj ) + tmp2 * fj;
        } else {
                tmp = -1.0;
        }
        return tmp;
}

/* see if we're below the mesh, e.g. we hit it. if so, return height at impact */
float detect_below(glheight_t *H, float *loc) {
        float tmp, fi, fj, h00, h01, h10, h11, tmp1, tmp2;
        int i, j;

	/* check if point *loc is in bounding rectangle of heightfield */
        if ((loc[0] < H->bbox[0]) && (loc[0] > H->location[0]) &&
	    (loc[2] < H->bbox[2]) && (loc[2] > H->location[2]) &&
	    (loc[1] < H->bbox[1])) {
                tmp = ( loc[0] - H->location[0] ) / (H->bbox[0] - H->location[0]);
                i = tmp * H->xdim;
                fi = tmp * (float) H->xdim - i;
                tmp = ( loc[2] - H->location[2] ) / (H->bbox[2] - H->location[2]);
                j = tmp * H->ydim;
                fj = tmp * (float) H->ydim - j;
                h00 = H->heights[j * H->xdim + i];
                h01 = H->heights[j * H->xdim + i + 1];
                h10 = H->heights[(j+1) * H->xdim + i];
                h11 = H->heights[(j+1) * H->xdim + i + 1];
                /* bilinear interpolate height */
                tmp1 = h00 * (1.0 - fi ) + h01 * fi;
                tmp2 = h10 * (1.0 - fi ) + h11 * fi;
                tmp = tmp1 * (1.0 - fj ) + tmp2 * fj;
		if (tmp < loc[1])  /* above the mesh... */
			tmp = -1.0;
	} else {
		tmp = -1.0;
	}
	return tmp;
}

void print_heightfield(glheight_t *H) {

	printf(
	"float location[3]; %3.2f %3.2f %3.2f\n"
	"float scale_factors[3];  %3.2f %3.2f %3.2f\n"
	"float bbox[3]; %3.2f %3.2f %3.2f\n"
	"int	xdim, ydim; %d %d\n"
	"float *heights;\n"
	"char *heightfname, *texfname; %s, %s\n"
	"unsigned int dlist, texid; %d, %d\n",
	H->location[0],
	H->location[1],
	H->location[2],
	H->scale_factors[0], 
	H->scale_factors[1], 
	H->scale_factors[2], 
	H->bbox[0],
	H->bbox[1],
	H->bbox[2],
	H->xdim, H->ydim,
	H->heightfname, H->texfname,
	H->dlist, H->texid);
}
