/* This code was created by Jeff Molofee '99 (Ported to Linux/GLX by
 * Patrick Schubert 2000 with the help from lesson 1 basecode for Linux/GLX
 * by Mihael Vrbanec).
 *
 * Cleaned up and Hopefully easier to read version by Jason Schultz (2004)
 *
 * @File:			glx_base.c
 *
 * @Description: 	Main Source File for OpenGL/GLX Base code
 */
 
#include "glx_base.h"
 
#include "player.h"
#include "gameeng.h"
#include "grapheng.h"
 
/* Define our window structure here
 *
 * Description of Structure Members: */
typedef struct
{
 
	Display 		*dpy;		/* This is the display. In X Windows, a display considers
 							everything. Monitor, Mouse, Keyboard. */
	int	  		screen;		/* Which window we are rendering too ? */
	Window 			win;		/* This is our drawable surrface */
	XWindowAttributes 	win_attr;	/* window attributues gotten, like resolution */
	GLXContext 		ctx;		/* Stores information about our OpenGL Context */
	XSetWindowAttributes	attr;		/* window attributes to set, like event types  */
	BOOL			fs;		/* This is the fullscreen mode flag */
	XF86VidModeModeInfo 	deskMode;	/* resolution of default desktop */
	int			x,y;		/* window position */
	unsigned int		width;		/* Width of the window */
	unsigned int		height;		/* Height of the window */
	unsigned int 		bpp;		/* Bits Per Pixel. With XFree86, highest we can acheive is 24.*/
	BOOL			warp;
	int 			warpx, warpy;	/* center of window, to warp pointer */
}GLWindow;

static GLWindow GLWin;	/* Set our OpenGL Window to static, we only want one */

/* createGLWindow(const char* title, int width, int height, int bpp, BOOL fsflag)
 *
 * Here we create the XFree86 window, with a GLX context.
 */
void createGLWindow(const char* title, int width, int height, int bpp, BOOL fsflag)
{
	int attrListSgl[] = {GLX_RGBA, GLX_RED_SIZE, 4,
					GLX_GREEN_SIZE, 4,
					GLX_BLUE_SIZE, 4,
					GLX_DEPTH_SIZE, 16,
					None};

	int attrListDbl[] = {GLX_RGBA, GLX_DOUBLEBUFFER,
					GLX_RED_SIZE, 4,
					GLX_GREEN_SIZE, 4,
					GLX_BLUE_SIZE, 4,
					GLX_DEPTH_SIZE, 16,
					None};

	XVisualInfo *vi;
	Colormap cmap;
	Cursor nullcursor;
	int dpyWidth, dpyHeight;
	int i;
	int vidModeMajorVersion, vidModeMinorVersion;
	XF86VidModeModeInfo **modes;
	int modeNum;
	int bestMode;
	Atom wmDelete;
	Window winDummy;
	unsigned int borderDummy;
	int deskres_x, deskres_y, winres_x, winres_y, win_x, win_y; /* desktop res, window res, window coord */

	GLWin.fs = fsflag;
	bestMode = 0;

	GLWin.dpy = XOpenDisplay(0); /* get conenction to X server */
	GLWin.screen = DefaultScreen(GLWin.dpy);
	XF86VidModeQueryVersion(GLWin.dpy, &vidModeMajorVersion,
							&vidModeMinorVersion);

	XF86VidModeGetAllModeLines(GLWin.dpy, GLWin.screen, &modeNum, &modes);

	/* save desktop modeline, has desktop resolution. for return from fullscreen */
	GLWin.deskMode = *modes[0];

	/* position window in center of screen */
	deskres_x = GLWin.deskMode.hdisplay;
	deskres_y = GLWin.deskMode.vdisplay;
	winres_x = width;
	winres_y = height;
	win_x = (deskres_x - winres_x) / 2;
	win_y = (deskres_y - winres_y) / 2;

	for (i = 0; i < modeNum; i++)
	{
		/*		printf("Mode %d: [ %d x %d ] flags: %x \n", i, modes[i]->hdisplay, modes[i]->vdisplay, modes[i]->flags); */
		if ((modes[i]->hdisplay == width) && (modes[i]->vdisplay == height))
		{
			bestMode = i;
		}
	}

	/* try to get double buffered screen */
	vi = glXChooseVisual(GLWin.dpy, GLWin.screen, attrListDbl);
	if(NULL == vi)
	{
		/* failing that, just assume you can get a single buffered one */
		vi = glXChooseVisual(GLWin.dpy, GLWin.screen, attrListSgl);
		printf("Singlebuffered : true\n");
	}
	else
	{
		printf("Doublebuffered : true\n");
	}

	GLWin.ctx = glXCreateContext(GLWin.dpy, vi, 0, GL_TRUE);

	/* color maps are vestigal.  Just make one and don't allocate any actual color cells */
	cmap = XCreateColormap(GLWin.dpy, RootWindow(GLWin.dpy, vi->screen),vi->visual, AllocNone);
	GLWin.attr.colormap = cmap;
	GLWin.attr.border_pixel = 0;

	/* Override redirect means "don't manage my window", e.g. no title bar, frame, etc. */
	GLWin.attr.override_redirect = False;
	
	if(GLWin.fs)
	{
		/* FULL SCREEN MODE: override redirect, vidmode changes */
		XF86VidModeSwitchToMode(GLWin.dpy, GLWin.screen, modes[bestMode]);
		XF86VidModeSetViewPort(GLWin.dpy, GLWin.screen, 0, 0);
		dpyWidth = modes[bestMode]->hdisplay;
		dpyHeight = modes[bestMode]->vdisplay;
		XFree(modes);
		GLWin.attr.override_redirect = True;
		GLWin.attr.event_mask = ExposureMask | KeyPressMask | KeyReleaseMask | ButtonPressMask |  
							StructureNotifyMask | PointerMotionMask | ButtonReleaseMask;
		GLWin.win = XCreateWindow(GLWin.dpy, RootWindow(GLWin.dpy, vi->screen),
								0, 0, dpyWidth, dpyHeight, 0, vi->depth, InputOutput, vi->visual,
								CWBorderPixel | CWColormap | CWEventMask | CWOverrideRedirect,
								&GLWin.attr);
		XWarpPointer(GLWin.dpy, None, GLWin.win, 0, 0, 0, 0, 0, 0);
		XGetWindowAttributes( GLWin.dpy, GLWin.win, &GLWin.win_attr);

		GLWin.warpx = GLWin.win_attr.width / 2; /* warp to middle of window in fullscreen mode */
		GLWin.warpy = GLWin.win_attr.height / 2;
		GLWin.warp = 1;
		XWarpPointer(GLWin.dpy, None, GLWin.win, 0, 0, 0, 0, GLWin.warpx, GLWin.warpy);
		XMapRaised(GLWin.dpy, GLWin.win);
		XGrabKeyboard(GLWin.dpy, GLWin.win, True, GrabModeAsync,GrabModeAsync, CurrentTime);
		nullcursor = createnullcursor( GLWin.dpy, GLWin.win);
		XGrabPointer(GLWin.dpy, GLWin.win, True, ButtonPressMask,
		GrabModeAsync, GrabModeAsync, GLWin.win, nullcursor, CurrentTime);
	}
	else
	{
		/* WINDOW MODE: get kill-window atom WM_DELETE_WINDOW */
		GLWin.attr.event_mask = ExposureMask | KeyPressMask | KeyReleaseMask | ButtonPressMask |
								StructureNotifyMask | ButtonMotionMask | ButtonReleaseMask;
		GLWin.win = XCreateWindow(GLWin.dpy, RootWindow(GLWin.dpy, vi->screen),
								  win_x, win_y, width, height, 0, vi->depth, InputOutput, vi->visual,
								  CWBorderPixel | CWColormap | CWEventMask | CWOverrideRedirect, &GLWin.attr);
		GLWin.warpx = 0; /* don't warp pointer in window mode */
		GLWin.warpy = 0;
		GLWin.warp = 0;
		wmDelete = XInternAtom(GLWin.dpy, "WM_DELETE_WINDOW", True);
		XSetWMProtocols(GLWin.dpy, GLWin.win, &wmDelete, 1);
		XSetStandardProperties(GLWin.dpy, GLWin.win, title,
		title, None, NULL, 0, NULL);
		XMapRaised(GLWin.dpy, GLWin.win);
       		/* XSetInputFocus(GLWin.dpy, GLWin.win, RevertToNone, CurrentTime); only for override redirect */
	} /* endif GLWin.fs */

	glXMakeCurrent(GLWin.dpy, GLWin.win, GLWin.ctx);
	XGetGeometry(GLWin.dpy, GLWin.win, &winDummy, &GLWin.x, &GLWin.y,
				&GLWin.width, &GLWin.height, &borderDummy, &GLWin.bpp);

	printf("Direct Rendering: %s\n",glXIsDirect(GLWin.dpy, GLWin.ctx) ? "true" : "false");

	printf("Running in %s mode\n",GLWin.fs ? "fullscreen" : "window");

	/* set viewport and projection matrix */
	resizeGLScene();
	initKeys(0); /* keyboard event configuration. */

}

/* runGLScene(void)
 *
 * Main loop for the program. while 1 [ process X events. draw gl scene. swap buffers ]
 */
/* event loop notes:
Keyboard: set autorepeat on/off:
	XAutoRepeatOn(GLWin.dpy); XAutoRepeatOff(GLWin.dpy);
	reduces keyboard events while holding key down. Unless you're
	entering text, this is a good thing 
	Keys are translated.  See keyboard module
Mouse: either requre a button down for mouse events, or
	use XQuery... to periodically poll mouse.  Warping
	the mouse allows for infinite lateral movement in
	any direction.  It's probably a good idea to hide the pointer */
void mainLoop(game_t *G)
{
/*	unsigned char ascii_key; */
	player_t *ego;	/* pointer to user's player structure */
	XEvent event;
	KeySym key_symbol;

	/* XQueryPointer parameters */
 	Window root_return, child_return;
        int root_x_return, root_y_return, win_x_return, win_y_return;
	int idx;
	stopwatch_t *sw;
	float mouse_event_time, event_time, submit_time, update_time, draw_time, swapbuffers_time, total_time;
	int event_counter;
        unsigned int mask_return;
	char keybuffer[10]; /* 10 keys down at once maximum. seemed like
				a lot at the time. */
	int running = 1, throttle = 100;

	sw = &G->sw;


	ego = &(G->players[G->first_person_index]);  
	printf("my ego is %d\n", G->first_person_index);
	player_print(ego); 

	for ( idx = 0; idx < 9 ; idx++) 
		keybuffer[idx] = 0x01;
	keybuffer[idx] = 0x00;
	
	stopwatch(sw, 0); /*reset stopwatch*/
	while(running)
	{
		event_counter = 0;
		/* mouse movement */
		mouse_event_time = stopwatch(sw, 2);
		if (! --throttle) {
			throttle = 1;
 			XQueryPointer(GLWin.dpy, GLWin.win, &root_return, &child_return, &root_x_return,
				&root_y_return, &win_x_return, &win_y_return, &mask_return);
			if (GLWin.warp != 0) {
				gr_motion(ego, win_x_return - GLWin.warpx , win_y_return - GLWin.warpy);
				XWarpPointer(GLWin.dpy, None, GLWin.win, 0, 0, 0, 0, GLWin.warpx, GLWin.warpy);
			} else {
				gr_motion(ego, win_x_return - GLWin.warpx , win_y_return - GLWin.warpy);
				GLWin.warpx = win_x_return;
				GLWin.warpy = win_y_return;
			}
		}
		event_time = stopwatch(sw, 2);
		/* check for X events */
		while(XPending(GLWin.dpy) > 0)
		{
			event_counter++;
			/* load &event with event */
			XNextEvent(GLWin.dpy, &event);
			switch(event.type)
			{
				case Expose:
					if (event.xexpose.count != 0)
						break;
					break;
				case ConfigureNotify:
					if ((event.xconfigure.width != GLWin.width) ||
						(event.xconfigure.height != GLWin.height))
					{
						GLWin.width = event.xconfigure.width;
						GLWin.height = event.xconfigure.height;
						printf("Resize Event\n");
						resizeGLScene();
					}
					break;
				case KeyPress:
					key_symbol = XLookupKeysym(&event.xkey,0);
					switch(key_symbol)
					{
						case XK_Escape:		/* Quit application */
							running = 0;
							break;
						case XK_F1:										
							/* Switch between fullscreen and windowed mode */
							killGLWindow();
							GLWin.fs = !GLWin.fs;
							createGLWindow("Nehe's GLX Base Code", 1280, 1024, 24, GLWin.fs);
							/* we lost our context, so re-init */
							gr_init(G);
							break;
            					default:  /* anything else - check if it is a letter key */
							genKeyString(&event.xkey, keybuffer);
                					break;
						}
						break;
				case KeyRelease:
					switch(XLookupKeysym(&event.xkey,0))
					{
						default:
							genKeyString(&event.xkey, keybuffer);
                					break;
					}
					break;
				case ClientMessage:
					/*Freaky following line simply checks first letter of atom name against "W" */
					printf("Atom name %s\n", XGetAtomName(GLWin.dpy, event.xclient.message_type));
					printf("Atom sizes: atom %d, ptr to atom %d\n", sizeof(Atom), sizeof( Atom * ));
					printf("SubAtomic name %s at %ld\n", XGetAtomName(GLWin.dpy, (Atom) event.xclient.data.l[0]),
						event.xclient.data.l[1]);
					printf("SubAtomic name %x %x %x %x %x %x %x %x\n",
						event.xclient.data.b[0],
						event.xclient.data.b[1],
						event.xclient.data.b[2],
						event.xclient.data.b[3],
						event.xclient.data.b[4],
						event.xclient.data.b[5],
						event.xclient.data.b[6],
						event.xclient.data.b[7]);
					if (*XGetAtomName(GLWin.dpy, event.xclient.message_type) == *"WM_PROTOCOLS")
					{
						running = 0;
					}
					break;
				case MotionNotify:
					if (GLWin.warpx == 0)
						gr_motion(ego, event.xbutton.x, event.xbutton.y);
					break;
				case ButtonPress:
					printf("Button pressed: %x\n", event.xbutton.button);
					gr_mouse(ego, event.xbutton.button, 1,  event.xbutton.x, event.xbutton.y);
					break;
				case ButtonRelease:
					printf("Button released: %x\n", event.xbutton.button);
					gr_mouse(ego, event.xbutton.button, 0, event.xbutton.x, event.xbutton.y);
					break;
				default:
					break;
			}
		}
		
		gr_keyboard(ego, keybuffer); 
		/* network time */
		game_submit(G, ego); 
		submit_time = stopwatch(sw, 2);
		game_update(G); 
		update_time = stopwatch(sw, 2);
		/* player_print(ego); */
		/* game engine time */
		gr_draw(G, ego);
		draw_time = stopwatch(sw, 2);
		swapBuffers();
		swapbuffers_time = stopwatch(sw, 2);
		total_time = stopwatch(sw, 1);
		stopwatch(sw, 0);
		G->loop_time = total_time;
		G->net_in_time = submit_time;
		G->net_out_time = update_time;
		G->event_time = event_time;
		G->draw_time = draw_time;
		G->loop_counter++;
		/*if (G->loop_counter % 220 == 0) {
			printf("%8dClient stats: %3.2f / sec: Total time: %3.2f , Mouse: %3.2f Event: %3.2f, submit:  %3.2f, update: %3.2f, draw: %3.2f. swapB: %3.2f  Event count: %d\n",
				G->loop_counter,
			       10.0 / total_time, total_time * 10.0,
	           mouse_event_time / total_time, 
			 event_time / total_time,
			submit_time / total_time,
			update_time / total_time,
			swapbuffers_time / total_time,
			  draw_time / total_time, event_counter);
		} */
	}
}

/* killGLWindow(void)
 *
 * Destroys our window and OpenGL Context. Closes the display
 */
void killGLWindow(void)
{
	if(GLWin.ctx)
	{
		if(!glXMakeCurrent(GLWin.dpy, None, NULL))
		{
			printf("Error releasing drawing context : killGLWindow\n");
		}
		glXDestroyContext(GLWin.dpy, GLWin.ctx);
		GLWin.ctx = NULL;
	}

	if(GLWin.fs)
	{
		XF86VidModeSwitchToMode(GLWin.dpy, GLWin.screen, &GLWin.deskMode);
		XF86VidModeSetViewPort(GLWin.dpy, GLWin.screen, 0, 0);
	}
	initKeys(1); /* keyboard event configuration. */
	XCloseDisplay(GLWin.dpy);
}

/* swapBuffers(void)
 *
 * Swaps the buffers (easy enough) ;o)
 */
void swapBuffers(void)
{
	glXSwapBuffers(GLWin.dpy, GLWin.win);
}

/* resizeGLScene(void)
 *
 * Resizes the window
 */
void resizeGLScene(void)
{
	/* float width,height; */
	if(GLWin.height == 0)
		GLWin.height = 1;
	if(GLWin.width == 0)
		GLWin.width = 1;
/*	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glViewport(0,0,GLWin.width,GLWin.height);
	width = 0.5f;
	height = 0.5f * ((float)(GLWin.height)/GLWin.width);
	glFrustum(-width,width,-height,height,1.0f,100.0f); or 
	glOrtho(0.0, 1.0, 0.0, 1.0, -1.0, 2.0);
	glMatrixMode(GL_MODELVIEW); */
	gr_reshape(GLWin.width,GLWin.height);
}

void initKeys(char autorepeat) /* keyboard event configuration. */
{
	if (autorepeat)
		XAutoRepeatOn(GLWin.dpy); 
	else
		XAutoRepeatOff(GLWin.dpy);
}	

void genKeyString(XKeyEvent *kbd_event, char *keylist)
/* generate key string.  key string will have all keys currently
held down, in *keylist.  */
{
	int keylist_size;
	char *p, c;
	char done;
	p = keylist;
	
	done = 0;

	XLookupString(kbd_event, &c, 1, NULL, NULL);

	if (c == 0x0) {
		printf("kbd_event translated to NULL:%d\n", (int) kbd_event);
		return;
	}
	while (*p) {
		if (*p == c) /* if already down, clear if up or down event */
			*p = 0x01; /* no-key code */
		else if (*p == 0x01 && kbd_event->type == KeyPress && !done) {
			*p = c;
			done = 1;
		}	
		p++;
	}
	/* keylist is full, so hash it in */
	if (kbd_event->type == KeyPress && !done) {
		keylist_size = (p - keylist);
		keylist[c % keylist_size] = c;
	}
		
}


/*
 * based on createnullcursor() from unclutter: Mark M Martin. cetia feb 1994  mmm@cetia.fr
 * UNCLUTTER-CODE-START
 */
Cursor createnullcursor( Display *display, Window root)
{
    Pixmap cursormask;
    XGCValues xgc;
    GC gc;
    XColor dummycolour;
    Cursor cursor;

    cursormask = XCreatePixmap(display, root, 1, 1, 1/*depth*/);
    xgc.function = GXclear;
    gc =  XCreateGC(display, cursormask, GCFunction, &xgc);
    /* XFillRectangle(display, cursormask, gc, 0, 0, 1, 1); */
    XDrawPoint(display, cursormask, gc, 0, 0);
    dummycolour.pixel = 0;
    dummycolour.red = 0;
    dummycolour.flags = 04; /* DoRed */
    cursor = XCreatePixmapCursor(display, cursormask, cursormask,
	      &dummycolour,&dummycolour, 0,0);
    XFreePixmap(display,cursormask);
    XFreeGC(display,gc);
    return cursor;
}
/* UNCLUTTER CODE END */
