#include <stdio.h>
#include <string.h>
#include <math.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glext.h>

#include "gameeng.h"
#include "grapheng.h"
#include "glprintf.h"
#include "glheightfield.h"

void setTexParams(void)
{
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
        /*glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE); */
        /* leave as modulate
        glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);  */
}


void gr_init(game_t *G) {
	/*char *hfn = "blurmaze.ppm";
	char *htn = "cblurmaze.ppm"; */

	float alpha = 0.5;

	printf("gr init\n");

	glClearColor(0.0, 0.1, 0.1, 0.0);
        glEnable(GL_DEPTH_TEST);
        glShadeModel(GL_SMOOTH);

	/* set up fonts */
	G->fontList = makefont(); 
	/* terrain */
/*	Now in config file. 
	G->heightfield.heightfname = hfn;
	G->heightfield.texfname = htn;
	G->heightfield.scale_factors[0] = 200.0;
	G->heightfield.scale_factors[1] = 15.0;
	G->heightfield.scale_factors[2] = 200.0;
	G->heightfield.location[0] = 0.0;
	G->heightfield.location[1] = 0.0;
	G->heightfield.location[2] = 0.0; */
	

	glEnable(GL_TEXTURE_2D);
	setTexParams();
	make_height_gr(&G->heightfield);
	/* printf("hf info:");
	print_heightfield(&G->heightfield); */

        /*glBlendColor(0.0, 0.0, 0.0, alpha);
        glBlendFunc(GL_CONSTANT_ALPHA, GL_ONE_MINUS_CONSTANT_ALPHA);
        glBlendEquation(GL_FUNC_ADD); */
}

void gr_keyboard(player_t *p, char *keybuffer) {
	int keymask;
	char *c;

	if (p == NULL)
		return;
	
	keymask = 0;
	c = keybuffer;

        while (*c) {
		if (*c != 0x01)
		switch (*c) {
			case 'w': keymask |= PLAYER_KEY_FORWARD; break;  /*forward*/
                        case 's': keymask |= PLAYER_KEY_BACKWARD; break; /* back */
                        case 'a': keymask |= PLAYER_KEY_STRAFE_LEFT; break;  /* strafe left */
                        case 'd': keymask |= PLAYER_KEY_STRAFE_RIGHT; break; /* strafe right */
                        case 'z': keymask |= PLAYER_KEY_LEFT; break;  /* rotate left */
                        case 'x': keymask |= PLAYER_KEY_RIGHT; break; /* rotate right */
			default:
                                printf("No key bound for '%c' dec: %d hex: %x\n",*c,*c,*c);
                        break;
		}
                c++;
        }
	p->key_actions = keymask;
}

void gr_mouse(player_t *p, int but, int state, int x, int y) {
	if (p == NULL)
		return;
        switch (but) {
               case 1:
                       if (state == 0) {
                               p->mouse_buttons &= ( ! PLAYER_MOUSE_BUTTON_0);
                       } else {
                               /* fire!! */
                               p->mouse_buttons |= PLAYER_MOUSE_BUTTON_0;
                       }
			break;
               case 2:
                       if (state == 0) {
                               p->mouse_buttons &= ( ! PLAYER_MOUSE_BUTTON_1);
                       } else {
                               p->mouse_buttons |= PLAYER_MOUSE_BUTTON_1;
                       }
			break;
               case 3:
                       if (state == 0) {
                               p->mouse_buttons &= ( ! PLAYER_MOUSE_BUTTON_2);
                       } else {
                               p->mouse_buttons |= PLAYER_MOUSE_BUTTON_2;
                       }
			break;
	}

}

void gr_motion(player_t *p, int x, int y) {
	if (p == NULL)
		return;
	p->mouse_delta[0] = x;
	p->mouse_delta[1] = y;
}

void gr_drawPlayer(player_t *p) {
	float radius = 1.5;

	glPushMatrix();
	glTranslatef(p->location[0],p->location[1],p->location[2]);
     	glRotatef(-p->heading[1], 0.0, 1.0, 0.0);
        glRotatef(p->heading[0], 1.0, 0.0, 0.0);

        glBegin(GL_TRIANGLE_FAN);
                glColor3ub(0x00, 0xff, 0xff);
                glVertex3f(0.0,  radius * 0.25 , 00);
                glColor3ub(0x00, 0x00, 0xff);
                glVertex3f(radius, 0.0, 0.);
                glVertex3f(0.0, 0.0,  radius  / 3.0 );
                glVertex3f(-radius, 0.0, 0.);
                glVertex3f(0.0, 0.0, -radius * 3.0);
                glVertex3f(radius, 0.0, 0.);
        glEnd();
        glBegin(GL_TRIANGLE_FAN);
                glColor3ub(0x0, 0x8f, 0xff);
                glVertex3f(0.0,  -radius * 0.25 , 00);
                glColor3ub(0xcf, 0xff, 0x00);
                glVertex3f(radius, 0.0, 0.);
                glVertex3f(0.0, 0.0,  radius / 3.0  );
                glVertex3f(-radius, 0.0, 0.);
                glVertex3f(0.0, 0.0, -radius * 3.0);
                glVertex3f(radius, 0.0, 0.);
        glEnd();
        glPopMatrix();

}
void gr_drawFPS(game_t *G, float x, float y ) {
                                                                                
        static char msg[40];
        static float fps = 0.0;
        float partial_fps;
	float timedelta;

	timedelta = G->loop_time;
                                                                                
        /* accumulate average fps over 10 frames */
        if (timedelta > 0.0)  {
                partial_fps = 10.0 / timedelta;
                fps = (fps * 9.0 + partial_fps) * 0.1;
        }
                                                                                
                                                                                
        glColor3ub(0x00, 0xff, 0x80);
        sprintf(msg, "FPS: %6.2f", fps);
        glPrintf(G->fontList, msg, x, y);
}

 
void gr_drawCrosshair() {
                glPushMatrix();
        glMatrixMode(GL_PROJECTION);
        glPushMatrix();
        glLoadIdentity();
        glOrtho(0.0, 1.0, 0.0, 1.0, -1.0, 1.0);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
        /* glRasterPos2i(x, y); */
        glColor3f(0.8, 0.3, 0.3);
        glBegin(GL_LINE_LOOP);
        glVertex2f(0.49, 0.49);
        glVertex2f(0.49, 0.51);
        glVertex2f(0.51, 0.51);
        glVertex2f(0.51, 0.49);
        glEnd();
        glMatrixMode(GL_PROJECTION);
        glPopMatrix();
        glMatrixMode(GL_MODELVIEW);
        glPopMatrix();
}


void gr_drawFire(bullet_t *file_list) {
	printf("drawfire\n");
}

void gr_drawBullets(bullet_t *b) {
	float radius, f, x, y, z;
                                                                                                                           
	while (b) {
        	radius = 0.17;
        	glPushMatrix();
	
        	/* orient viewer */
        	glTranslatef(b->location[0],b->location[1],b->location[2]);
		 switch (b->type) {
                        case BULLET_TYPE_ROCKET:
        			glRotatef(-b->heading[1], 0.0, 1.0, 0.0);
        			glRotatef(b->heading[0], 1.0, 0.0, 0.0);
        			/* draw pyramid in gray */
        			glColor3ub(0x99, 0xf9, 0x99);
        			glBegin(GL_TRIANGLE_FAN);
                			glVertex3f(0.0, 0.0, -1.0);
                			glColor3ub(0x9f, 0x99, 0xf9);
                			glVertex3f(radius, 0.0, -0.5);
                			glVertex3f(0.0, radius,  -0.5);
                			glVertex3f(-radius, 0.0, -0.5);
                			glVertex3f(0.0, -radius, -0.5);
                			glVertex3f(radius, 0.0, -0.5);
        			glEnd();
				break;
                        case BULLET_TYPE_EXPLOSION:
        glBlendColor(0.0, 0.0, 0.0, 0.5);
        glBlendFunc(GL_CONSTANT_ALPHA, GL_ONE_MINUS_CONSTANT_ALPHA);
        glBlendEquation(GL_FUNC_ADD);
				glEnable(GL_BLEND);
				radius = b->heading[0];
        			glBegin(GL_TRIANGLE_FAN);
                			glColor3ub(0xa0, 0xa0, 0x00);
                			glVertex3f(0.0, 0.0, -radius);
                			glColor3ub(0xa0, 0x80, 0x00);
                			glVertex3f(radius, 0.0, 0.0);
                			glVertex3f(0.0, radius,  0.0);
                			glVertex3f(-radius, 0.0, 0.0);
                			glVertex3f(0.0, -radius, 0.0);
                			glVertex3f(radius, 0.0, 0.0);
        			glEnd();
        			glBegin(GL_TRIANGLE_FAN);
                			glColor3ub(0xa0, 0xa0, 0x00);
                			glVertex3f(0.0, 0.0, radius);
                			glColor3ub(0xa0, 0x80, 0x00);
                			glVertex3f(radius, 0.0, 0.0);
                			glVertex3f(0.0, radius,  0.0);
                			glVertex3f(-radius, 0.0, 0.0);
                			glVertex3f(0.0, -radius, 0.0);
                			glVertex3f(radius, 0.0, 0.0);
        			glEnd();
				glDisable(GL_BLEND);
				break;
                        case BULLET_TYPE_SHIP_EXPLOSION:
				radius = b->heading[0];
				glBegin(GL_TRIANGLES);
				for (f = 0.0 ; f < 120.0; f+=1.0) {
					x = (sin(f) ) * radius;
					y = (sin(f * 0.3 + radius * 0.8 + 0.6)  ) * radius;
					z = (sin(f * 1.52 + radius * 41.2 + 1.6)  ) * radius;
					if ( (int) f % 3 == 0)
						glColor3ub(0xff, (unsigned char) (radius * 951.0) % 0x80, 0x00);
					else
						glColor3ub(0xff,  0x00, 0x00);
					glVertex3f(x, y, z);
				}
				glEnd();
				break;
		}

        	glPopMatrix();
		b = b->next;
	}
}

void gr_drawLand_orig() {
	int i, j, k;
	/* sky */
	glColor3ub(0x10, 0x60, 0x90);
	glBegin(GL_QUADS);
	glVertex3f(-100.0, 65.0, -100.0);	
	glVertex3f(100.0, 65.0, -100.0);	
	glVertex3f(100.0, 65.0, 100.0);	
	glVertex3f(-100.0, 65.0, 100.0);	
	glEnd();
	/* ground */
	glColor3ub(0x10, 0xa0, 0x10);
	glBegin(GL_QUADS);
	glVertex3f(-100.0, -65.0, -100.0);	
	glVertex3f(100.0, -65.0, -100.0);	
	glVertex3f(100.0, -65.0, 100.0);	
	glVertex3f(-100.0, -65.0, 100.0);	
	glEnd();

	/* stuff */
	glPushMatrix();
	for ( i = -10 ; i < 10 ; i += 4)
		for ( j = -10 ; j < 10 ; j += 4)
			for ( k = -10 ; k < 10 ; k += 4) {
				glColor3ub(0x10 * i, 0x10 * j, 0x10 * k);
				glTranslatef((float) i * 5.0, (float) j, (float) k * 5.0);
				glBegin(GL_TRIANGLE_FAN);
					glVertex3f(0.0, 1.0, 0.0);
					glVertex3f(1.0, -1.0, 0.0);
					glVertex3f(0.0, -1.0, 1.0);
					glVertex3f(-1.0, -1.0, 0.0);
					glVertex3f(0.0, -1.0, -1.0);
					glVertex3f(1.0, -1.0, 0.0);
				glEnd();
				glTranslatef((float) -i * 5.0, (float) -j, 
								(float) -k * 5.0);
			}
	glPopMatrix();
}

void gr_drawLand(game_t *G) {
	float lx, ly, lz, bx, by, bz;

	lx =	G->heightfield.location[0];
	ly =	G->heightfield.location[1];
	lz =	G->heightfield.location[2];
	bx =	G->heightfield.bbox[0];
	by =	G->heightfield.bbox[1];
	bz =	G->heightfield.bbox[2];

	/*printf("%3.2f %3.2f %3.2f = %3.2f %3.2f %3.2f \n",
	lx, ly, lz, bx, by, bz); */
	/* print_heightfield(&G->heightfield); */

	glColor3ub(0x10, 0x60, 0x90);
	glBegin(GL_QUADS);

	glVertex3f( bx, ly, lz );
	glVertex3f( bx, by, lz );
	glVertex3f( bx, by, bz );
	glVertex3f( bx, ly, bz );

	glVertex3f( lx, by, lz );
	glVertex3f( lx, by, bz );
	glVertex3f( lx, ly, bz );
	glVertex3f( lx, ly, lz );
	glEnd();

	glColor3ub(0xff, 0xff, 0xff);
	draw_height_gr(&G->heightfield); 

	/* the ocean! */
	glColor3ub(0x10, 0xa0, 0xf0);
	glBegin(GL_QUADS);
	glVertex3f(lx, 0.2, lz);
	glVertex3f(lx, 0.2, bz);
	glVertex3f(bx, 0.2, bz);
	glVertex3f(bx, 0.2, lz);
	glEnd();
}

void gr_drawStats(game_t *G)
{
	float tt;

	char pfmt[] = "#P:%2d[M:%4d/t:%5.2f p:%d, s:%d ps:%d pps:%d]";
	char pfmt2[] = "#C:%2d Tm:%5.2f >:%5.2f <:%5.2f E:%5.2f D:%5.2f";
	char msg[256];

	tt = 100.0 / G->loop_time;

	glColor3ub(0xff, 0xff, 0x00);
	sprintf(msg, pfmt, G->loop_packets,
			G->max_packet_size,
			G->max_packet_wait,
			G->max_packet_b,
			G->max_packet_s,
			G->max_packet_ps,
			G->max_packet_pps);
	glPrintf(G->fontList, msg, 0.0, 0.1); 
	sprintf(msg, pfmt2,
			G->client_game_cycle_delta[G->first_person_index],
			G->loop_time,
			G->net_in_time * tt,
			G->net_out_time * tt,
			G->event_time * tt,
			G->draw_time * tt
			); 
	glPrintf(G->fontList, msg, 0.0, 0.05); 
}

void gr_draw(game_t *G, player_t *p) {
	int i;
	static const char *hp_fmt = "Hit Points: %4d";
	char hp_msg[256];
	
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
                                                                                
        glColor3f(0.0, 1.0, 0.0);
        /* orient viewer */
        glLoadIdentity();
        glRotatef(-p->heading[0], 1.0, 0.0, 0.0);
        glRotatef(p->heading[1], 0.0, 1.0, 0.0);
        glTranslatef(-p->location[0], -p->location[1], -p->location[2]);
        glTranslatef(0.0,-p->height,0.0);

	/* draw land */
	gr_drawLand(G);
	gr_drawBullets(G->bullet_list_start);

	/* draw players */
	for (i = 0 ; i < GAME_MAX_CLIENTS; i++) {
		if (i != G->first_person_index)
			if (G->players[i].status == PLAYER_STATUS_ALIVE)
				gr_drawPlayer(&G->players[i]);
	}
	/* draw console */
	glColor3ub(0x00, 0xff, 0x00);
	sprintf(hp_msg, hp_fmt, G->players[G->first_person_index].hit_points); 
	glPrintf(G->fontList, hp_msg, 0.75, 0.0); 
	gr_drawCrosshair();
	gr_drawFPS(G, 0.8, 0.9 );
	gr_drawStats(G);
}

void gr_reshape(int w, int h) {
        GLfloat aspect;
                                                                                                                                                          
        aspect = (GLfloat) w / (GLfloat) h;
                                                                                                                                                          
        glViewport(0, 0, (GLsizei) w, (GLsizei) h);
        glMatrixMode(GL_PROJECTION);
        glLoadIdentity();
        /* gaspect = aspect; */
        gluPerspective(90.0, aspect, 0.01, 1000.0);
        glMatrixMode(GL_MODELVIEW);
        glLoadIdentity();
}


