#ifndef QUATERNION_H
#define QUATERNION_H

/* index by row,column or column,row? */
#ifndef COLUMN_ORDER
#ifndef ROW_ORDER
#define COLUMN_ORDER
#endif
#endif

#define QUATERNION_USE_SINGLE
/* use double or single precision? */
#ifndef QUATERNION_USE_DOUBLE
#define qFloat float
#else
#define qFloat double
#endif

typedef qFloat quaternion[4];
typedef	qFloat m16[16];
typedef	qFloat v1;
typedef	qFloat v2[2];
typedef	qFloat v3[3];
typedef	qFloat v4[4];

/* leave result in parameter list vars, typ. q1 */
qFloat	* qnew_acc(quaternion q1, qFloat theta, qFloat *u);
qFloat	* qmul_acc(quaternion q1, quaternion q2);
qFloat	* qadd_acc(quaternion q1, quaternion q2);
qFloat	* qcopy_acc(quaternion dest, quaternion src);
qFloat 	* qinv_acc(quaternion q);
qFloat   * qget_acc(quaternion q, qFloat *theta, qFloat *u);

/* return result in allocated memory */
qFloat	* qnew(qFloat theta, qFloat u[3]);
qFloat	* qmul(quaternion q1, quaternion q2);
qFloat	* qadd(quaternion q1, quaternion q2);
qFloat	* qcopy(quaternion q2);
qFloat 	* qinv(quaternion q);

/*util*/
void 	  qprint(quaternion q);
qFloat	* qrotate(quaternion q, qFloat *v);
qFloat	* q_to_matrix(quaternion q, m16 m);
qFloat	* matrix_to_q(quaternion q, m16 m);
qFloat	* lerp(qFloat *dest, qFloat *start, qFloat *end, qFloat alpha, qFloat len);
qFloat  * slerp_precomp(qFloat *sphi, qFloat *x, qFloat *y, int len);
qFloat	* slerp(qFloat *sphi, quaternion qdest, quaternion q1,quaternion q2, qFloat alpha);

/* vector: if returns *qFloat, then result is in 1st parameter */
qFloat     vec_lensq(qFloat *v, int len); /* equiv to vec_dot(v,v) */
qFloat	  vec_dot(qFloat *v1, qFloat *v2, int len);
qFloat	* vec_unit(qFloat *v, int len);
qFloat	* vec_scalar_mul(qFloat d, qFloat *v, int len);
qFloat	* vec_add(qFloat *v1, qFloat *v2, int len);
qFloat	* vec_copy(qFloat *dst, qFloat *src, int len);
qFloat 	* vec_cross(qFloat *v1, qFloat *v2, int len);
/* matrix */
qFloat	  determinant(qFloat *m, int n); /* M is NxN matrix */
qFloat	* mtranspose(qFloat *m, int w, int h);
qFloat	* adjoint(qFloat *m, int n);
qFloat	* minverse(qFloat *m, int n);
qFloat	* mcopy_acc(qFloat *mdst, qFloat *msrc, int w, int h);
qFloat	* mcopy(qFloat *msrc, int w, int h);
void	  mprint(qFloat *m, int w, int h);
qFloat 	* matrix_scalar_mul(qFloat f, qFloat *mtarg, int m, int n); 
qFloat 	* matrix_mul(qFloat *m1, qFloat *m2, int m1m, int m1n, int m2m, int m2n);

/* supply allocated *cofmat  */
qFloat 	* cofactor_matrix(qFloat *cofmat, qFloat *srcmat,int i,int j , int n);
qFloat 	  cofactor(qFloat *cofmat, int i, int j, int n);
float deg2rad(float deg);

#endif /* QUATERNION_H */
